@@ -14,7 +14,11 @@
 #include <a/fncdcls.h>
 
 /* external macro declarations */
+#if (defined(__sgi) && _MIPS_SZLONG == 64) || defined(__sparcv9)
+#define MD      63
+#else
 #define MD	31
+#endif
 
 /* external data declarations */
 /* number of words in block of given scale */
