@@ -4,8 +4,7 @@
 /* See .../src/LICENSE for terms of distribution.                           */
 /*                                                                           */
 /*                                                                           */
-/*****************************************************************************/
-#include <stdio.h>
+/*****************************************************************************/#include <stdio.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
@@ -17,13 +16,14 @@
 
 #include <dap/balloc.h>
 
-#include <a/development.h>
+#include <development.h>
 
 #include <a/f.h>
 #include <a/fncdcls.h>
 #include <a/ik.h>
 #include <a/arthur.h>
 #include <a/fir.h>
+#include <a/beam.h>
 
 extern I dbg_tb, dbg_tnan;
 
@@ -31,24 +31,10 @@ extern C *findMapped32FileName();
 extern C *findMapped64FileName();
 extern C *findFileName();
 
-#if defined(__alpha) || defined(__ia64)
-char beamFileCPrefix[8]={'A','p','l','u','s','L','6','4'};
-int *beamFilePrefix=(int *)beamFileCPrefix;
-#elif (_MIPS_SZLONG == 64)
-char beamFileCPrefix[8]={'A','p','l','u','s','B','6','4'};
-int *beamFilePrefix=(int *)beamFileCPrefix;
-#elif defined(__VISUAL_C_2_0__) || defined(__i386)
-char beamFileCPrefix[8]={'A','p','l','u','s','l','3','2'};
-int *beamFilePrefix=(int *)beamFileCPrefix;
-#else
-int  *beamFilePrefix=0;
-#endif
-
-
-
-#define ENDIAN_UNDEF  0
-#define ENDIAN_LITTLE 1
-#define ENDIAN_BIG    2
+/* flag to control auto/endain conversion of the .m file  */
+Z I autoBeamConvert=0;		/* Default is off  */
+I getAutoBeamConvert(void)   {R autoBeamConvert;}
+void setAutoBeamConvert(I m) {autoBeamConvert=m;}
 
 Z I mapOut(C *s,A z) ;
 
@@ -268,12 +254,6 @@ Z int loopen(C *path,int flags,mode_t mode)
   R fd;
 }
 
-struct a32 {
-    int c,t,r,n,d[MAXR],i,p[1];
-};
-
-#define AH32 (sizeof(struct a32) - sizeof(int))
-
 C *tmv32(int t,I *d,int *s,int n) {
   int i;
 
@@ -347,7 +327,7 @@ gc32(int t,int r,int n,int d[],int *p) {
 }
 
 A
-cvt64(struct a32 *aobj) {   /* mapped objects only - floats, chars and longs */
+cvt64(A32 *aobj) {   /* mapped objects only - floats, chars and longs */
     A newa;
 	int i = aobj->i ;
     /*printf("%d %d %d %d %d\n", aobj->c, aobj->t, aobj->r, aobj->n, aobj->i) ; */
@@ -356,8 +336,8 @@ cvt64(struct a32 *aobj) {   /* mapped objects only - floats, chars and longs */
     R newa;
 }
 
-#if defined(__alpha) || defined(__i386) || defined(__ia64)
-Z A vetteMappedFile(I fsize,I aarg,int mode,C *name)
+
+Z A vetteMappedFile(I fsize,I aarg,int mode,C *name, I *rc, I cvtInPlace)
 {
   A aobj=(A)aarg,z;
 
@@ -365,128 +345,9 @@ Z A vetteMappedFile(I fsize,I aarg,int mode,C *name)
 /*   return  1, conversion successful, z is new object in memory */
 /*   return -1, conversion failed,     z is set to NULL */
 
-  if( 1==cvtIfNeeded(aobj, &z, fsize) )	/* Conversion Successful */
-    1; /* need to unmap aobj */
-  
+  *rc=cvtIfNeeded(aobj, &z, fsize, cvtInPlace);
   return z;
-
 }
-#else
-Z A vetteMappedFile(I fsize,I aarg,int mode,C *name)
-{
-  A aobj=(A)aarg,z=(A)0;
-  static struct a tempheader;
-  /*printf("In vetteMappedFile(%ld, %ld, %d, %s)\n", 
-		fsize, aarg, mode, name) ; */
-  if (aarg!=-1 && 0==aobj->c)
-  {
-    if(isWrongEndian(aobj))
-    {
-      if(ENDIAN_UNDEF!=aobjEndian(aobj))
-      {
-	ndnicopy((I *)aobj,(I *)&tempheader,AH/sizeof(I));
-	if(Ct>=tempheader.t&&fsize>=AH+tempheader.n)
-	{
-	  if(BEAM_RO==mode)
-	  {
-	    H("%s[warn]: Beam: File %s is wrong endian.  Making local copy\n",
-	      CC,name);
-	    z=ep_CopyRightEndian(aobj);
-	    dc(aobj);
-	  }
-	  else
-	  {
-	    H("%s[warn]: Beam: File %s is wrong endian.  Aborting map.\n",
-	      CC,name);
-	    z=0;
-	  }
-	}
-      }
-    }
-    else if(Ct>=aobj->t&&fsize>=AH+Tt(aobj->t,aobj->n))z=aobj;
-  }
-  if((I)aobj == -1)
-	{
-	printf("\343 Internal error vetteMappedFile(): This is being doubly converted. Bad.\n") ;
-	return (A)0 ;
-	}
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
-  if(!z)
-	{
-	I m, t, items ;
-	int ret, fd;
-	char * dotPos ;
-
-	printf("\343 Converting 32 bit file.\n") ; 
-    z=cvt64((struct a32 *)aobj);
-	t = z->t ;
-  	 /* printf("z->c: %ld z->t: %ld z->r: %ld z->n: %ld z->d[0]: %ld z->d[1]: %ld z->i: %ld\n",
-			z->c, t, z->r, z->n, z->d[0], z->d[1], z->i) ; 
-	*/
-    if (z==NULL)
-	  {
-	  H("%s[warn]: Beam error: not an `a object\n",CC);
-      z = NULL;
-	  }
-	/* ok it's converted. now we can unlink */
-	/* printf("Creating 64 bit mapped file (.m64)\n") ; */
-	dotPos = strrchr(name, '.') ;
-	if (dotPos)
-		*dotPos = 0 ;
-	name=findFileName(name,"m64");
-
-	/* printf("Attempting to write the file out to %s\n", name) ;  */
-	fd=loopen(name,O_CREAT|O_WRONLY,0666) ;
-	if (!fd)
-		{
-		printf("\343 Error in loopen(%s)\n", name) ;
-		return (A)0 ;
-		}
-	/* we hacked the i correctly already so ignore d[0] */
-	z->c = 0 ;
-	items = z->i ;
-	if (z->n > z->i)
-		items = z->n ;
-	ret=writeAobjToFile(fd,(C *)z,AH+Tt(t,items)+(Ct==t?1:0),1);
-	if (ret < 0)
-		{
-		printf("\343 Error in writeAobjToFile: %d\n", ret) ;
-		return (A)0 ;
-		}
-	m=z->i*tr(z->r-1,z->d+1);
-	if(z->i<*z->d)
-	  {
-	  *z->d=z->i ;
-	  z->n=m;
-	  }
-	/* printf("m = %ld\n", m) ; */
-	/* printf("Setting file len to %ld\n", AH+Tt(t,m)+(t==Ct)) ; */
-	/* if(-1==flen(fd,AH+Tt(t,m)+(t==Ct)))
-		{
-		perror(name) ;
-		close(fd);
-		q=9 ;
-		return (A)0;
-		}
-	*/
-/*	if(-1==lseek(fd,0,0))  R perror(name),close(fd),q=9,(A)0;
-	if(-1==write(fd,(C *)z,AH))R perror(name),close(fd),q=9,(A)0; */
-
-  	 /* printf("z->c: %ld z->t: %ld z->r: %ld z->n: %ld z->d[0]: %ld z->d[1]: %ld z->i: %ld\n",
-			z->c, t, z->r, z->n, z->d[0], z->d[1], z->i) ; */
-
-	printf("\343 Wrote %s.\n", name) ;
-	close(fd) ;
-	dc(z) ; /* This should free */
-	return (A)-1 ;
-	}
-#else
-	if (!z)
-		H("%s[warn]: Beam error: not an `a object\n",CC);
-#endif
-  R z;
-}
-#endif
 
 typedef struct{I a,c,n,w;C *s,*t;}MFInfo;  
   /* n field used for "next" in freelist */
@@ -596,7 +457,7 @@ extern I map(int,int);
 Z I mapDotMFile(int fd,int mode,C *fname,C *t)
 {
   A aobj;
-  I mfile;
+  I mfile,rc;
   MFInfo *p;
   I n;
   (void) setStickyBit(fd,0);
@@ -604,8 +465,59 @@ Z I mapDotMFile(int fd,int mode,C *fname,C *t)
   mfile=map(fd,mode);		/* map returns a 0 on failure    */
   Q(!mfile,9);			/* Set domain error if map fails */
   /* printf("mfile = %ld, n = %ld\n", mfile, n) ; */
-  aobj=vetteMappedFile(n,mfile,mode,fname);
+  aobj=vetteMappedFile(n, mfile, mode, fname, &rc,
+		       mode==BEAM_RW && autoBeamConvert);
+  if(rc==-1) 
+    {
+      munmap((caddr_t)mfile,n);
+/*       H("\343 Conversion error\n"); */
+      R q=9,0;
+    }
+
   Q(!aobj,9);
+
+  /* The following if clause will convert the .m file to the native format */
+  /* for 32 bit to 64 bit otherwise it is done in place */
+
+  if(rc==1 && mode==BEAM_RW && autoBeamConvert==1) /* Convert .m file */
+    {
+      struct a a;
+      I itemCount=aobj->i;
+      if(dbg_tb) printf("\343 Converting %s\n",t);
+      p=getFreeMFInfoStruct();
+      if(!p)
+        {
+          R H("maplim\n"),dc(aobj),q=9,0;
+        }
+      p->c=1,p->n=n,p->w=mode,p->s=bstring(fname),p->t=bstring(t),p->a=mfile;
+      unmapDotMFile((A)mfile,p);
+
+#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__sparcv9)
+      t = findMapped64FileName(fname, mode) ;
+      if (!t)
+        t = findMapped32FileName(fname, mode) ;
+#endif
+
+      mapOut(t, aobj);
+      dc(aobj);
+      aobj=0;
+
+      ERR(t, fd=loopen(t,O_RDWR,0666));
+      if(-1==read(fd,(C *)&a,AH))
+	return perror(t), close(fd), q=9, 0;
+      if(-1==flen(fd,AH+Tt(a.t,a.i)+(a.t==Ct)))
+	return perror(t), close(fd), q=9, 0;
+      if(-1==lseek(fd,0,SEEK_SET))
+	return perror(t), close(fd), q=9, 0;
+
+      setStickyBit(fd,0);
+      mfile=map(fd,mode);               /* map returns a 0 on failure    */
+      Q(!mfile,9);                      /* Set domain error if map fails */
+      if(dbg_tb)beamtrc(t,1,mode);
+      aobj=vetteMappedFile(n,mfile,mode,fname,&rc,0);
+      Q(!aobj,9);
+    }
+
   p=getFreeMFInfoStruct();
   if(!p)
   {
@@ -706,15 +618,11 @@ I mapIn(C *name,I mode){
   C*t=0;
   /* printf("In mapIn(%s, %d)\n", name, mode) ;*/
   Q(!name||BEAM_RO>mode||BEAM_LOCAL<mode,9);
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
+#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__sparcv9)
   /* First look for a 64 bit file */
   t = findMapped64FileName(name, (BEAM_RW==tmode)) ;
   if (!t)
-	{
-	/* We are only reading */
-	tmode = BEAM_RW ;
-	t = findMapped32FileName(name, (BEAM_RW==tmode)) ;
-	}
+    t = findMapped32FileName(name, (BEAM_RW==tmode)) ;
 #else
   t = findMapped32FileName(name, (BEAM_RW==tmode)) ;
 #endif
@@ -722,39 +630,21 @@ I mapIn(C *name,I mode){
   if(dbg_tb)beamtrc(t,1,tmode);
   ERR(t,fd=loopen(t,BEAM_RW==tmode?O_RDWR:O_RDONLY,0666));
   z=mapDotMFile(fd, tmode, name, t) ;
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
-  switch(z)
-	{
-	case 0:
-	  R 0 ;
-	case -1:
-	  /* printf("Converted 32 bit file, mapping file again\n") ; */
-	  close(fd) ;
-	  t = findMapped64FileName(name, (BEAM_RW==mode)) ;
-	  /* printf("t = %s\n", t) ; */
-	  ERR(t,fd=loopen(t,BEAM_RW==mode?O_RDWR:O_RDONLY,0666));
-	  /* printf("Calling mapDotMFile(%d, %d, %s, %s)\n",
-			fd, mode, name, t) ; */
-	  if(0==(z=mapDotMFile(fd,mode,name,t)))R 0;
-	  break ;
-	}
-#endif
-  if(dbg_tnan)nanbeamchk(t,(A)z);
-  /* printf("Returning z = %ld\n", (I)z) ; */
+  if(z && dbg_tnan )nanbeamchk(t,(A)z);
   R z;
 }
 
 Z writeAobjToFile(int fd,C *s,unsigned n,I c)
 {
-  I t;
+  I t, newItems;
   A a=(A) s; 
   /* printf("in writeAobj - n = %u\n", n) ; */
   do t=write(fd,s,n); while(s+=t,t!=-1&&(n-=t));
-  if((!c) && t!=-1 && a->i!=a->d[0]) 
+  if((!c) && t!=-1 && a->i!=(newItems=a->r ? a->d[0] : a->n)) 
   { 
     /* if beaming out a mapped variable, writeover a->i with a->d[0]  */
     t=lseek(fd,(AH-sizeof(I)),SEEK_SET);
-    if(t!=-1) t=write(fd,(C *)a->d,4);
+    if(t!=-1) t=write(fd,(C *)&newItems,sizeof(I));
   }
   fsync(fd);R t;
 } /* IBM write fix */
@@ -764,7 +654,7 @@ Z I mapOut(C *s,A z)
   I fd,rc,c=0;
   C r[MAXPATHLEN];
   C *dotPos, *slashPos;
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
+#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__sparcv9)
   static C defaultSuffix[]={"m64"} ;
 #else
   static C defaultSuffix[]={"m"};
@@ -843,71 +733,177 @@ Z I items(I n,A z)
   C *name ;
   Q(-2>n,9);
   NDC1(z);s=stringFromAobj(z);Q(!s,9);
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
+
+#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__sparcv9)
   name = findMapped64FileName(s, w) ;
   if (!name)
 	name = findMapped32FileName(s, w) ;
 #else
   name = findMapped32FileName(s, w) ;
 #endif
-  if(-1==(fd=loopen(name,w?O_RDWR:O_RDONLY,0666))) 
-    return q=9,0;
+
+  ERR(s, fd=loopen(name,w?O_RDWR:O_RDONLY,0666));
   if(-1==read(fd,(C *)&a,AH)) 
     return perror(s),close(fd),q=9,0;
-#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__ia64)
+
   {
     I ret, totsize,rank,itemCount,items;
     struct stat statbuf;
-    ret = fstat(fd, &statbuf) ;
-    if (ret)
-    {
-      close(fd);
-      perror("items (fstat):") ;
-      return 0;
-    }
+    if ( -1==fstat(fd, &statbuf) ) /* To get size for getItems() */
+      return perror(s), close(fd), q=9, 0;
     
     ret=getItems(&a, &itemCount, &rank, &items, statbuf.st_size);
     
-    /* domain error if conversion error or */
-    /* conversion and trying to set        */
-    if( ret==-1 || (ret && n!=-1) )
+    if( ret==-1 || rank==0 )	/* error conversion or scalar */
       {
-	 close(fd);
-	 q=9;
-	 return 0;
+	printf("\343 Error: %s [%s]\n", 
+	       (ret==-1) ? "Conversion ":"Scalar ",name);
+	return close(fd), q=(ret==-1)?9:7, 0;
       }
-    else if( ret==1 )		/* Conversion was required     */
+
+    if( n==-1 )			/* Query */
+      return close(fd), itemCount>items?itemCount:items;
+
+    if( ret==1 && !autoBeamConvert ) /* conversion required */
       {
-	close(fd);
-	Q( !rank, 7);		/* Must be not be a scalar */
-	return itemCount>items?itemCount:items;
+	printf("\343 Error: requires conversion [%s]\n",name);
+	return close(fd), q=9, 0;
       }
-  }
-#endif		
-  if(!a.r) R close(fd),q=7,0;
-  m=*a.d;j=a.i;if(m>j)j=m;
-  if(w){
-    t=a.t;
-    if(n==-2) 
-    {
-      if(-1==setSizeOfFile(fd,AH+Tt(t,a.n)+(t==Ct)))
-	R perror(s),close(fd),q=9,0;
-    }
-    else
-    {
-      a.i=n,m=n*tr(a.r-1,a.d+1);if(n<*a.d)*a.d=n,a.n=m;
-      if(-1==flen(fd,AH+Tt(t,m)+(t==Ct))) R perror(s),close(fd),q=9,0;
-      if(-1==lseek(fd,0,0))  R perror(s),close(fd),q=9,0;
-      if(-1==write(fd,(C *)&a,AH))R perror(s),close(fd),q=9,0;
+
+    /* try to set the items */
+    if( ret==1 )		/* convert before setting items */
+      {
+	A aobj=(A)mapDotMFile(fd, BEAM_RW, s, name);
+	if(aobj==0) 
+	  return 0;
+	else
+	  dc(aobj);
+#if (_MIPS_SZLONG == 64) || defined(__alpha) || defined(__sparcv9)
+	name = findMapped64FileName(s, w) ;
+	if (!name)
+	  name = findMapped32FileName(s, w) ;
+#endif
+				/* re-open converted file */
+	if(-1==(fd=loopen(name,w?O_RDWR:O_RDONLY,0666))) 
+	  return perror(s),q=9,0;
+	if(-1==read(fd,(C *)&a,AH)) 
+	  return perror(s),close(fd),q=9,0;
+      }
+
+    m=*a.d;			/* Set the items */
+    j=a.i;
+    if(m>j)j=m;
+    
+    if(w){
+      t=a.t;
+      if(n==-2)
+	{
+	  if(-1==setSizeOfFile(fd,AH+Tt(t,a.n)+(t==Ct)))
+	    R perror(s),close(fd),q=9,0;
+	}
+      else
+	{
+	  a.i=n;
+	  m=n*tr(a.r-1,a.d+1);
+	  if(n<*a.d)	/* reset d[0] if greater than n */
+	    *a.d=n,a.n=m;
+	  if(-1==flen(fd,AH+Tt(t,m)+(t==Ct))) 
+	    R perror(s),close(fd),q=9,0;
+	  if(-1==lseek(fd,0,0))  
+	    R perror(s),close(fd),q=9,0;
+	  if(-1==write(fd,(C *)&a,AH))
+	    R perror(s),close(fd),q=9,0;
+	}
     }
+    R close(fd),j;
   }
-  R close(fd),j;
 }
 
+static long ep_isaDotM(char *fname)
+{
+  long rc=0;                    /* 0 == invalid dot M file */
+  A aobj;
+  I mlen,fd;
+
+  if(-1==(fd=open(fname,O_RDONLY)))
+    {
+      perror(fname);
+      return rc;
+    }
+
+  mlen=lseek(fd,0,2);
+  if(0==(aobj=(A)map(fd,0)))    /* map closes fd */
+    {
+      perror(fname);
+      return rc;
+    }
+
+  if ( isAObject(*aobj) )
+    {
+      int dimsOK=1;
+      /* Check dimensions */
+      checkDims(*aobj);
+
+      /* Check file length */
+#if (HAS_64BIT_TYPE==1)
+      if (dimsOK && checkFileSize64(*aobj) <= mlen)
+        rc=1;
+#else
+      if (dimsOK && checkFileSize32(*aobj) <= mlen)
+        rc=1;
+#endif
+    }
+
+  munmap(aobj,mlen);
+  return rc;
+}
+
+static void ep_msyncAll(A msyncMode_)
+{
+  int msyncFlag;
+  long i;
+
+  if(msyncMode_->t!=Et || !QS(*msyncMode_->p) )
+    {
+      q=ERR_DOMAIN;
+      R;
+    }
+
+  if(*msyncMode_->p==MS(si("MS_ASYNC")))
+    {
+      msyncFlag=MS_ASYNC;
+    }
+  else if(*msyncMode_->p==MS(si("MS_SYNC")))
+    {
+      msyncFlag=MS_SYNC; 
+    }
+  else
+    {
+      q=ERR_DOMAIN;
+      R;
+    }
+
+  for(i=0; i<maxMFAindex; i++)
+    {
+      MFInfo *p=MappedFileArray+i;
+      if(p->a  && p->w==1 )
+	{
+	  if(-1==msync((A)p->a, p->n, msyncFlag))
+	    {
+	      printf("\343 Error: %s\n",p->t); 
+	      perror("ep_amsyncAll: msync");
+	    }
+	}
+    }
+} 
+
+
 void beamInstall(void){
   install((PFI)items,        "_items",     9,2,9,0,0,0,0,0,0,0);
   install((PFI)ep_hostEndian,"_hostendian",0,0,0,0,0,0,0,0,0,0);
   install((PFI)ep_aobjEndian,"_endian",    0,1,0,0,0,0,0,0,0,0);
+  install((PFI)ep_isaDotM,   "_isaDotM",   9,1,7,0,0,0,0,0,0,0);
+  install((PFI)ep_msyncAll,  "_msyncAll",  8,1,0,0,0,0,0,0,0,0);
   R;
 }
 
