@@ -428,6 +428,7 @@ Z C *SysVarList[] = { "dyme", "vers", "pp", "mode", "stop", "Df", "Gf",
 		      "releaseCode", "language", "Ef", "si", "segvexit",
 		      "busexit", "loadfile", "Xfpef", "maplim",
 		      "doErrorStack","CCID",
+                      "autoBeamConvert","beamMSyncMode",
 		      (char *)0 };
 
 Z C *ModeList[] = { "ascii", "apl", "uni", (char *)0 };
@@ -517,6 +518,8 @@ A ep_gsv(A aname)
     CSR(26, R (A)gi(MFALimitGet())); /* maplim */
     CSR(27, R (A)gi(doErrorStack));
     CSR(28, R CCID ? gsym(CCID):aplus_nl);
+    CSR(29, R (A)gi(getAutoBeamConvert()));
+    CSR(30, R getBeamMSyncMode());              /* beamMSyncMode */
   default: ERROUT(ERR_DOMAIN);
   }
 }
@@ -596,6 +599,9 @@ A ep_ssv(A aname, A aval)
     CS(25, RtCHK(aval, 1); xfpeFlag=ival;);
     CS(26, ItCHK(aval);MFALimitSet(ival););
     CS(27, RtCHK(aval, 1); doErrorStack=ival;);
+    CSR(28, ERROUT(ERR_DOMAIN));               /* CCID */
+    CS(29,  ItCHK(aval);setAutoBeamConvert (ival););
+    CS(30, setBeamMSyncMode(aval));            /* beamMSyncMode */
   default: ERROUT(ERR_DOMAIN);
   }
   R 0;
