@@ -76,7 +76,6 @@ extern fsync(),getsockopt(),setsockopt();
  *
  * Error Code Table and Functions
  */
-extern int errno;
 
 #if !defined(__osf__)
 
@@ -403,15 +402,30 @@ A cpu()
   R r;
 }
 
-I amsync(a, flags)
-  A a;
-  I flags;
+#ifndef _AIX
+static MaskTable MsyncMasks[] = {
+
+  { MS_ASYNC,     "MS_ASYNC",     0,      0 },
+  { MS_SYNC,      "MS_SYNC",      0,      0 },
+  { MS_INVALIDATE,"MS_INVALIDATE",0,      0 },
+  { 0,            (char *)0,      0,      0 }
+};
+#endif
+
+I amsync(A a, A aflags)
 {
 #ifdef _AIX
   R 0;
 #else
   int bytes;
 
+  unsigned long flags;
+
+  if (SymbolsToMask(MsyncMasks, aflags, &flags) == -1) {
+    (void)pa((V)aflags);
+    q = ERR_DOMAIN;
+    R 0;
+  }
   bytes=mf_length(a);
   R bytes?msync((caddr_t)a, bytes, (int)flags):0;
 #endif /* _AIX */
@@ -1388,7 +1402,7 @@ void eponymousInstall()
   install((PFI)sysaccess, "access", IV,2,A_,A_,0,0,0,0,0,0);
   install((PFI)agetdents, "agetdents", A_, 1, CP,0,0,0,0,0,0,0);
   install((PFI)alstat, "alstat", A_, 1, CA,0,0,0,0,0,0,0);
-  install((PFI)amsync,"amsync", IV,2,A_,IV,0,0,0,0,0,0);
+  install((PFI)amsync,"amsync", IV,2,A_,A_,0,0,0,0,0,0);
   install((PFI)areadlink, "areadlink", A_, 1, CP,0,0,0,0,0,0,0);
   install((PFI)aselect, "aselect", A_,4,IA,IA,IA,IA,0,0,0,0);
   install((PFI)astat, "astat", A_, 1, CA,0,0,0,0,0,0,0);
